import os
import json
import yaml
import logging
import signal
import psutil
from argparse import Namespace
from pathlib import Path
from dgbowl_schemas.tomato import to_payload
from .. import setlib
from .. import dbhandler


log = logging.getLogger(__name__)


def submit(args: Namespace) -> None:
    """
    Job submission function. Usage:

    .. code:: bash

        ketchup [-t] [-v] [-q] submit <payload>

    Attempts to open the ``yaml/json`` file specified in the ``<payload>`` argument,
    and submit it to tomato's queue.

    The supplied :class:`argparse.Namespace` has to contain the path to the ``payload``.
    Optional arguments include the verbose/quiet switches (``-v/-q``) and the testing
    switch (``-t``).

    Examples
    --------

    >>> # Submit a job:
    >>> ketchup submit .\dummy_random_2_0.1.yml
    jobid = 2

    >>> # Increased verbosity:
    >>> ketchup -v submit .\dummy_random_2_0.1.yml
    INFO:tomato.ketchup.functions:Output path not set. Setting output path to 'C:\[...]'
    INFO:tomato.ketchup.functions:queueing 'payload' into 'queue'
    INFO:tomato.dbhandler.sqlite:inserting a new job into 'state'
    jobid = 4

    """
    dirs = setlib.get_dirs(args.test)
    settings = setlib.get_settings(dirs.user_config_dir, dirs.user_data_dir)
    queue = settings["queue"]
    assert os.path.exists(args.payload)
    assert os.path.isfile(args.payload)

    log.debug(f"attempting to open payload at '{args.payload}'")
    with open(args.payload, "r") as infile:
        if args.payload.endswith("json"):
            pldict = json.load(infile)
        elif args.payload.endswith("yml") or args.payload.endswith("yaml"):
            pldict = yaml.full_load(infile)
    payload = to_payload(**pldict)
    log.debug("payload=Payload(%s)", payload)
    if payload.tomato.output.path is None:
        cwd = str(Path().resolve())
        log.info("Output path not set. Setting output path to '%s'", cwd)
        payload.tomato.output.path = cwd
    pstr = payload.json()
    log.info("queueing 'payload' into 'queue'")
    jobid = dbhandler.queue_payload(queue["path"], pstr, type=queue["type"])
    print(f"jobid = {jobid}")


def status(args: Namespace) -> None:
    """
    Job, queue and pipeline status query function. Usage:

    .. code:: bash

        ketchup [-t] [-v] [-q] status
        ketchup [-t] [-v] [-q] status [queue|state]
        ketchup [-t] [-v] [-q] status <jobid>

    The :class:`argparse.Namespace` has to contain the ``<jobid>`` the status of
    which is supposed to be queried. Alternatively, the status of the ``queue``
    or ``state`` of tomato can be queried. Optional arguments include the verbose/
    quiet switches (``-v/-q``) and the testing switch (``-t``).

    Examples
    --------

    >>> # Get pipeline status of tomato:
    >>> ketchup status
    pipeline             ready jobid  (PID)     sampleid
    ===================================================================
    dummy-10             no    3      1035      dummy_sequential_1_0.05
    dummy-5              no    None             None

    >>> # Get queue status with queued & running jobs:
    >>> ketchup status queue
    jobid  status (PID)     pipeline
    ==========================================
    3      r      1035      dummy-10
    4      q

    >>> # Get queue status with all jobs:
    >>> ketchup -v status queue
    jobid  status (PID)     pipeline
    ==========================================
    1      c
    2      cd
    3      r      1035      dummy-10
    4      q

    >>> # Get status of a given job
    >>> ketchup status 1
    jobid = 1
    status = c
    submitted at = 2022-06-02 06:49:00.578619+00:00
    executed at  = 2022-06-02 06:49:02.966775+00:00
    completed at = 2022-06-02 06:49:08.229213+00:00

    """
    dirs = setlib.get_dirs(args.test)
    settings = setlib.get_settings(dirs.user_config_dir, dirs.user_data_dir)
    state = settings["state"]
    queue = settings["queue"]

    if args.jobid == "state":
        pips = dbhandler.pipeline_get_all(state["path"], type=state["type"])
        print(
            f"{'pipeline':20s} {'ready':5s} {'jobid':6s} {'(PID)':9s} {'sampleid':20s} "
        )
        print("=" * 67)
        for pip in pips:
            sampleid, ready, jobid, pid = dbhandler.pipeline_get_info(
                state["path"], pip, state["type"]
            )
            rstr = "yes" if ready else "no"
            job = f"{str(jobid):6s} ({pid})" if jobid is not None else str(jobid)
            print(f"{pip:20s} {rstr:5s} {job:16s} {str(sampleid):20s}")
    elif args.jobid == "queue":
        jobs = dbhandler.job_get_all(queue["path"], type=queue["type"])
        running = dbhandler.pipeline_get_running(state["path"], type=state["type"])
        print(f"{'jobid':6s} {'status':6s} {'(PID)':9s} {'pipeline':20s}")
        print("=" * 42)
        for jobid, payload, status in jobs:
            if status.startswith("q"):
                print(f"{str(jobid):6s} {status}")
            elif status.startswith("r"):
                for pip, pjobid, pid in running:
                    if pjobid == jobid:
                        print(f"{str(jobid):6s} {status:6s} {str(pid):7s} {pip:20s}")
            elif status.startswith("c") and args.verbose - args.quiet > 0:
                print(f"{str(jobid):6s} {status:6s}")
    else:
        jobid = int(args.jobid)
        ji = dbhandler.job_get_info(queue["path"], jobid, type=queue["type"])
        payload, status, submitted_at, executed_at, completed_at = ji
        print(f"jobid = {jobid}")
        print(f"status = {status}")
        print(f"submitted at = {submitted_at}")
        if status.startswith("r") or status.startswith("c"):
            print(f"executed at  = {executed_at}")
            running = dbhandler.pipeline_get_running(state["path"], type=state["type"])
            for pipeline, pjobid, pid in running:
                if pjobid == jobid:
                    print(f"with pipeline = {pipeline}")
                    print(f"with PID = {pid}")
                    break
        if status.startswith("c"):
            print(f"completed at = {completed_at}")


def cancel(args: Namespace) -> None:
    """
    Job cancellation function. Usage:

    .. code:: bash

        ketchup [-t] [-v] [-q] cancel <jobid>

    The :class:`argparse.Namespace` has to contain the ``<jobid>`` of the job to be
    cancelled. Optional arguments include the verbose/quiet switches (``-v/-q``) and
    the testing switch (``-t``).

    Examples
    --------

    >>> # Cancel a job:
    >>> ketchup cancel 1

    .. warning::

        Cancelling a running job will generate a warning. Output FAIR data should be
        created as requested in the ``<payload>``.

    >>> # Cancel a running job:
    >>> ketchup cancel 1
    WARNING:tomato.ketchup.functions:cancelling a running job 1 with pid 17584

    .. note::

        Cancelling a completed job will do nothing.


    """
    dirs = setlib.get_dirs(args.test)
    settings = setlib.get_settings(dirs.user_config_dir, dirs.user_data_dir)
    state = settings["state"]
    queue = settings["queue"]
    jobid = int(args.jobid)
    jobinfo = dbhandler.job_get_info(queue["path"], jobid, type=queue["type"])
    status = jobinfo[1]
    log.debug(f"found job {jobid} with status '{status}'")
    if status in {"q", "qw"}:
        log.info(f"setting job {jobid} to status 'cd'")
        dbhandler.job_set_status(queue["path"], "cd", jobid, type=queue["type"])
    elif status == "r":
        running = dbhandler.pipeline_get_running(state["path"], type=state["type"])
        for pip, pjobid, pid in running:
            if pjobid == jobid:
                log.warning(f"cancelling a running job {jobid} with pid {pid}")
                proc = psutil.Process(pid=pid)
                for cp in proc.children():
                    if cp.name() in {"python", "python.exe"}:
                        for ccp in cp.children():
                            log.debug(
                                "sending SIGTERM to pid %d with name '%s'",
                                ccp.pid,
                                ccp.name(),
                            )
                            ccp.send_signal(signal.SIGTERM)


def load(args):
    dirs = setlib.get_dirs(args.test)
    settings = setlib.get_settings(dirs.user_config_dir, dirs.user_data_dir)
    state = settings["state"]

    log.debug(f"checking whether pipeline '{args.pipeline}' exists.")
    pips = dbhandler.pipeline_get_all(state["path"], type=state["type"])
    assert args.pipeline in pips, f"pipeline '{args.pipeline}' not found."

    log.info(f"loading sample '{args.sample}' into pipeline '{args.pipeline}'")
    dbhandler.pipeline_load_sample(
        state["path"], args.pipeline, args.sample, type=state["type"]
    )


def eject(args):
    dirs = setlib.get_dirs(args.test)
    settings = setlib.get_settings(dirs.user_config_dir, dirs.user_data_dir)
    state = settings["state"]

    log.debug(f"checking whether pipeline '{args.pipeline}' exists.")
    pips = dbhandler.pipeline_get_all(state["path"], type=state["type"])
    assert args.pipeline in pips, f"pipeline '{args.pipeline}' not found."

    log.debug(f"checking if pipeline '{args.pipeline}' is running.")
    sampleid, ready, jobid, pid = dbhandler.pipeline_get_info(
        state["path"], args.pipeline, state["type"]
    )

    if sampleid is not None:
        assert jobid is None and pid is None, (
            "cannot remove a sample" " from a running pipeline"
        )
        log.info(f"ejecting sample '{sampleid}' from pipeline '{args.pipeline}'")
        dbhandler.pipeline_eject_sample(
            state["path"], args.pipeline, type=state["type"]
        )
    else:
        log.info(f"pipeline '{args.pipeline}' is already empty")


def ready(args):
    dirs = setlib.get_dirs(args.test)
    settings = setlib.get_settings(dirs.user_config_dir, dirs.user_data_dir)
    state = settings["state"]

    log.debug(f"checking whether pipeline '{args.pipeline}' exists.")
    pips = dbhandler.pipeline_get_all(state["path"], type=state["type"])
    assert args.pipeline in pips, f"pipeline '{args.pipeline}' not found."

    log.debug(f"checking if pipeline '{args.pipeline}' is running.")
    sampleid, ready, jobid, pid = dbhandler.pipeline_get_info(
        state["path"], args.pipeline, state["type"]
    )

    if jobid is None and pid is None:
        log.info(f"marking pipeline '{args.pipeline}' as ready.")
        dbhandler.pipeline_reset_job(state["path"], args.pipeline, True, state["type"])
    else:
        log.warning(f"cannot mark pipeline as ready: job '{jobid}' is running.")



""" Bio-Logic OEM package python API.

This module provides simple functions that are general purpose,
and prone to be used in several modules.
"""

import os

# ------------------------------------------------------------------------------#


def class_name(obj):
    """Return the class name of an object."""
    name = type(obj).__name__
    return name


# ------------------------------------------------------------------------------#


def exception_brief(e, extended=False):
    """Return either a simple version of an exception, or a more verbose one."""
    brief = f"{class_name(e)}"
    if extended:
        brief += f" : {e}"
    return brief


# ------------------------------------------------------------------------------#


def warn_diff(msg, cmp):
    """Check a predicate (assert) or a mismatch, and on error print a message."""
    if type(cmp) is bool:
        not_ok = not cmp
        if not_ok:
            print(f"{msg} failed")
    else:
        not_ok = cmp[0] != cmp[1]
        if not_ok:
            print(f"{msg} {cmp}")
    return not_ok


# ------------------------------------------------------------------------------#


def error_diff(msg, cmp):
    """Check a predicate (assert) or a mismatch, and on error raise an exception."""
    if type(cmp) is bool:
        if not cmp:
            raise RuntimeError(f"{msg}")
    else:
        if cmp[0] != cmp[1]:
            raise RuntimeError(f"{msg} {cmp}")


# ------------------------------------------------------------------------------#


def prepend_path(path, filename):
    """Prepend a path to filename in case one is not already provided."""
    if path:
        segs = os.path.split(filename)
        if (segs[0] == "") and (filename != ""):
            filename = path + filename
    return filename


# ------------------------------------------------------------------------------#


def file_complete(filename, an_ext):
    """Append an extension to a filename unless the file already exists or if it already has one."""
    if not os.path.isfile(filename):
        root, ext = os.path.splitext(filename)
    if not ext:
        filename = root + an_ext
    return filename


# ------------------------------------------------------------------------------#


def pp_plural(nb, label, num=True, nothing=""):
    """Return a user friendly version of an ordinal and a label.

    num is used to force a number version,
    nothing is what to say if there is nothing
    """
    if nb == 0:
        if nothing:
            en_clair = f"{nothing}"
        else:
            en_clair = f"{0 if num else 'no'} {label}"
    elif nb == 1:
        en_clair = f"{1 if num else 'one'} {label}"
    else:
        en_clair = f"{nb} {label}s"
    return en_clair


# ==============================================================================#



from typing import Any
from importlib import metadata
import importlib
import argparse
import time
import multiprocessing
import subprocess
import os
import json
from datetime import datetime, timezone
import logging

from .logger_funcs import log_listener_config, log_listener, log_worker_config
from .yadg_funcs import get_yadg_preset
from .. import dbhandler


def tomato_job() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--version",
        action="version",
        version=f'%(prog)s version {metadata.version("tomato")}',
    )
    parser.add_argument(
        "jobfile",
        help="Path to a ketchup-processed payload json file.",
        default=None,
    )
    args = parser.parse_args()

    jobfolder, _ = os.path.split(os.path.abspath(args.jobfile))
    logfile = args.jobfile.replace(".json", ".log")

    logging.basicConfig(
        level=logging.DEBUG,
        format="%(asctime)s:%(levelname)-8s:%(processName)s:%(message)s",
        handlers=[logging.FileHandler(logfile, mode="a"), logging.StreamHandler()],
    )
    logger = logging.getLogger(__name__)

    logger.info("attempting to load jobfile '%s'", args.jobfile)
    with open(args.jobfile, "r") as infile:
        jsdata = json.load(infile)

    logger.debug("parsing data from jobfile")
    jobid = jsdata["jobid"]
    settings = jsdata["settings"]
    queue = settings["queue"]
    state = settings["state"]
    payload = jsdata["payload"]
    tomato = payload.get("tomato", {})
    pipeline = jsdata["pipeline"]
    pip = pipeline["name"]

    verbosity = tomato.get("verbosity", "INFO")
    loglevel = logging._checkLevel(verbosity)
    logger.debug("setting logger verbosity to '%s'", verbosity)
    logger.setLevel(loglevel)

    # pid = os.getpid()
    pid = os.getppid()  # On Windows, the parent is the tomato_job.exe

    logger.debug(f"assigning job '{jobid}' on pid '{pid}' into pipeline '{pip}'")
    dbhandler.pipeline_assign_job(state["path"], pip, jobid, pid, type=state["type"])
    dbhandler.job_set_status(queue["path"], "r", jobid, type=queue["type"])
    dbhandler.job_set_time(queue["path"], "executed_at", jobid, type=queue["type"])

    logger.info("handing off to 'driver_worker'")
    logger.info("==============================")
    ret = driver_worker(settings, pipeline, payload, jobid, logfile, loglevel)

    logger.info("==============================")

    output = tomato["output"]
    prefix = f"results.{jobid}" if output["prefix"] is None else output["prefix"]
    path = output["path"]
    logger.debug("output path is '%s'", path)
    if os.path.exists(path):
        logger.debug("path exists, making sure it's a folder")
        assert os.path.isdir(path)
    else:
        logger.debug("path does not exist, creating")
        os.makedirs(path)
    dgfile = os.path.join(path, f"{prefix}.json")
    logging.debug("creating a preset file '%s'", f"preset.{jobid}.json")
    preset = get_yadg_preset(payload["method"], pipeline)
    with open(f"preset.{jobid}.json", "w") as of:
        json.dump(preset, of)

    logging.info("running yadg to create a datagram in '%s'", dgfile)
    command = ["yadg", "preset", "-pa", f"preset.{jobid}.json", jobfolder, dgfile]
    logging.debug(" ".join(command))
    subprocess.run(command, check=True)
    logging.debug("removing the preset file '%s'", f"preset.{jobid}.json")
    os.unlink(f"preset.{jobid}.json")

    ready = tomato.get("unlock_when_done", False)
    if ret is None:
        logger.info("job finished successfully, setting status to 'c'")
        dbhandler.job_set_status(queue["path"], "c", jobid, type=queue["type"])
    else:
        logger.info("job was terminated, setting status to 'cd'")
        dbhandler.job_set_status(queue["path"], "cd", jobid, type=queue["type"])
        logger.info("handing off to 'driver_reset'")
        logger.info("==============================")
        driver_reset(settings, pipeline)
        logger.info("==============================")
        ready = False

    logger.debug(f"setting pipeline '{pip}' as '{'ready' if ready else 'not ready'}'")
    dbhandler.pipeline_reset_job(state["path"], pip, ready, type=state["type"])
    dbhandler.job_set_time(queue["path"], "completed_at", jobid, type=queue["type"])


def driver_api(
    driver: str,
    command: str,
    jobqueue: multiprocessing.Queue,
    logger: logging.Logger,
    address: str,
    channel: int,
    **kwargs: dict,
) -> Any:
    m = importlib.import_module(f"tomato.drivers.{driver}")
    func = getattr(m, command)
    return func(address, channel, jobqueue, logger, **kwargs)


def data_poller(
    driver: str,
    jq: multiprocessing.Queue,
    lq: multiprocessing.Queue,
    address: str,
    channel: int,
    device: str,
    root: str,
    loglevel: int,
    kwargs: dict,
) -> None:
    log_worker_config(lq, loglevel)
    log = logging.getLogger()
    pollrate = kwargs.pop("pollrate", 10)
    log.debug(f"in 'data_poller', {pollrate=}")
    cont = True
    previous = None
    while cont:
        ts, done, _ = driver_api(
            driver, "get_status", jq, log, address, channel, **kwargs
        )
        ts, nrows, data = driver_api(
            driver, "get_data", jq, log, address, channel, **kwargs
        )
        data["previous"] = previous
        previous = data["current"]
        while nrows > 0:
            isots = datetime.fromtimestamp(ts, tz=timezone.utc).isoformat()
            isots = isots.replace(":", "")
            fn = os.path.join(root, f"{device}_{isots}_data.json")
            log.debug(f"found {nrows} data rows, writing into '{fn}'")
            with open(fn, "w") as of:
                json.dump(data, of)
            ts, nrows, data = driver_api(
                driver, "get_data", jq, log, address, channel, **kwargs
            )
            data["previous"] = previous
            previous = data["current"]
        if done:
            cont = False
        else:
            time.sleep(pollrate)
    log.info(f"rejoining main thread")
    return


def driver_worker(
    settings: dict,
    pipeline: dict,
    payload: dict,
    jobid: int,
    logfile: str,
    loglevel: int,
) -> None:

    jq = multiprocessing.Queue(maxsize=0)

    log = logging.getLogger(__name__)
    log.setLevel(loglevel)
    log.debug("starting 'log_listener'")
    lq = multiprocessing.Queue(maxsize=0)
    listener = multiprocessing.Process(
        target=log_listener,
        name="log_listener",
        args=(lq, log_listener_config, logfile),
    )
    listener.start()
    log.debug(f"started 'log_listener' on pid {listener.pid}")

    root = os.path.join(settings["queue"]["storage"], str(jobid))
    jobs = []
    for vi, v in enumerate(pipeline["devices"]):
        log.info(f"device id: {vi+1} out of {len(pipeline['devices'])}")
        log.info(f"{vi+1}: processing device '{v['tag']}' of type '{v['driver']}'")
        drv, addr, ch, tag = v["driver"], v["address"], v["channel"], v["tag"]
        dpar = settings["drivers"].get(drv, {})
        pl = [item for item in payload["method"] if item["device"] == v["tag"]]
        smpl = payload["sample"]

        log.debug(f"{vi+1}: getting status")
        ts, ready, metadata = driver_api(drv, "get_status", jq, log, addr, ch, **dpar)
        log.debug(f"{ready=}")
        assert ready, f"Failed: device '{tag}' is not ready."

        log.debug(f"{vi+1}: starting payload")
        start_ts = driver_api(
            drv, "start_job", jq, log, addr, ch, **dpar, payload=pl, **smpl
        )
        metadata["uts"] = start_ts

        log.debug(f"{vi+1}: writing initial status")
        fn = os.path.join(root, f"{tag}_status.json")
        with open(fn, "w") as of:
            json.dump(metadata, of)
        kwargs = dpar
        kwargs.update({"pollrate": v.get("pollrate", 10)})
        log.info(f"{vi+1}: starting 'data_poller': every {kwargs['pollrate']}s")
        p = multiprocessing.Process(
            name=f"data_poller_{jobid}_{tag}",
            target=data_poller,
            args=(drv, jq, lq, addr, ch, tag, root, loglevel, kwargs),
        )
        jobs.append(p)
        p.start()
        log.info(f"{vi+1}: started 'data_poller' on pid {p.pid}")

    log.info("waiting for all 'data_poller' jobs to join")
    log.info("------------------------------------------")
    ret = None
    for p in jobs:
        p.join()
        log.debug(f"{p=}")
        if p.exitcode == 0:
            log.info(f"'data_poller' with pid {p.pid} closed successfully")
        else:
            log.critical(f"'data_poller' with pid {p.pid} was terminated")
            ret = 1

    log.info("-----------------------")
    log.info("quitting 'log_listener'")
    lq.put_nowait(None)
    listener.join()
    jq.close()
    return ret


def driver_reset(
    settings: dict,
    pipeline: dict,
) -> None:
    log = logging.getLogger(__name__)
    for vi, v in enumerate(pipeline["devices"]):
        log.info(f"device id: {vi+1} out of {len(pipeline['devices'])}")
        log.info(f"{vi+1}: processing device '{v['tag']}' of type '{v['driver']}'")
        drv, addr, ch, tag = v["driver"], v["address"], v["channel"], v["tag"]
        dpar = settings["drivers"].get(drv, {})

        log.debug(f"{vi+1}: resetting device")
        driver_api(drv, "stop_job", None, log, addr, ch, **dpar)

        log.debug(f"{vi+1}: getting status")
        ts, ready, metadata = driver_api(drv, "get_status", None, log, addr, ch, **dpar)
        assert ready, f"Failed: device '{tag}' is not ready."



import psutil
import os
import subprocess
import time
import json
import logging
from .. import dbhandler


def _find_matching_pipelines(pipelines: list, method: list[dict]) -> list[str]:
    req_names = set([item["device"] for item in method])
    req_capabs = set([item["technique"] for item in method])

    candidates = []
    for cd in pipelines:
        dnames = set([dev["tag"] for dev in cd["devices"]])
        if req_names.intersection(dnames) == req_names:
            candidates.append(cd)

    matched = []
    for cd in candidates:
        capabs = []
        for v in cd["devices"]:
            capabs += v["capabilities"]
        if req_capabs.intersection(set(capabs)) == req_capabs:
            matched.append(cd)

    return matched


def _pipeline_ready_sample(ret: tuple, sample: dict) -> bool:
    sampleid, ready, jobid, pid = ret
    if ready == 0:
        return False
    else:
        if sample["name"] == sampleid:
            return True
        else:
            return False


def main_loop(settings: dict, pipelines: dict, test: bool = False) -> None:
    log = logging.getLogger(__name__)
    qup = settings["queue"]["path"]
    qut = settings["queue"]["type"]
    stp = settings["state"]["path"]
    stt = settings["state"]["type"]
    while True:
        # check existing PIDs in state
        ret = dbhandler.pipeline_get_running(stp, type=stt)
        for pip, jobid, pid in ret:
            log.debug(f"checking PID of running job '{jobid}'")
            if psutil.pid_exists(pid) and "tomato_job" in psutil.Process(pid).name():
                log.debug(f"PID of running job '{jobid}' found")
                # dbhandler.job_set_status(queue, "r", jobid)
            else:
                log.debug(f"PID of running job '{jobid}' not found")
                dbhandler.pipeline_reset_job(stp, pip, False, type=stt)
                dbhandler.job_set_status(qup, "ce", jobid, type=qut)
                dbhandler.job_set_time(qup, "completed_at", jobid, type=qut)

        # check existing jobs in queue
        ret = dbhandler.job_get_all(qup, type=qut)
        for jobid, strpl, st in ret:
            payload = json.loads(strpl)
            if st in ["q", "qw"]:
                if st == "q":
                    log.info(f"checking whether job '{jobid}' can ever be matched")
                matched_pips = _find_matching_pipelines(pipelines, payload["method"])
                if len(matched_pips) > 0 and st != "qw":
                    dbhandler.job_set_status(qup, "qw", jobid, type=qut)
                log.debug(f"checking whether job '{jobid}' can be queued")
                for pip in matched_pips:
                    pipinfo = dbhandler.pipeline_get_info(stp, pip["name"], type=stt)
                    can_queue = _pipeline_ready_sample(pipinfo, payload["sample"])
                    if can_queue:
                        log.info(f"queueing job '{jobid}' on pipeline '{pip['name']}'")
                        dbhandler.pipeline_reset_job(stp, pip["name"], False, type=stt)
                        args = {
                            "settings": settings,
                            "pipeline": pip,
                            "payload": payload,
                            "jobid": jobid,
                        }
                        root = os.path.join(settings["queue"]["storage"], str(jobid))
                        os.makedirs(root)
                        jpath = os.path.join(root, "jobdata.json")
                        with open(jpath, "w") as of:
                            json.dump(args, of, indent=1)
                        cfs = subprocess.CREATE_NO_WINDOW
                        if not test:
                            cfs |= subprocess.CREATE_NEW_PROCESS_GROUP
                        subprocess.Popen(
                            ["tomato_job", str(jpath)],
                            creationflags=cfs,
                        )
                        break
        time.sleep(settings.get("main loop", 1))



""" Bio-Logic OEM package python API.

This module contains support functions when building technique parameters,
and decoding experiment records.

"""

from dataclasses import dataclass

from . import kbio_types as KBIO
from .tech_types import TECH_ID


@dataclass
class ECC_parm:
    """ECC param template"""

    label: str
    type_: type


# functions to build the technique ECC parameters (structure+contents)


def make_ecc_parm(api, ecc_parm, value=0, index=0):
    """Given an ECC_parm template, create and return an EccParam, with its value and optional index."""
    parm = KBIO.EccParam()
    # BL_Define<xxx>Parameter
    # .. value is converted to its proper type, which DefineParameter will use
    api.DefineParameter(ecc_parm.label, ecc_parm.type_(value), index, parm)
    return parm


def make_ecc_parms(api, *ecc_parm_list):
    """Create an EccParam array from an EccParam list, and return an EccParams refering to it."""
    nb_parms = len(ecc_parm_list)
    parms_array = KBIO.ECC_PARM_ARRAY(nb_parms)

    for i, parm in enumerate(ecc_parm_list):
        parms_array[i] = parm

    parms = KBIO.EccParams(nb_parms, parms_array)
    return parms


# function to handle records from a running experiment


def print_experiment_data(api, data):
    """Unpack the experiment data, decode it according to the technique, display it,
    then return the experiment status"""

    current_values, data_info, data_record = data

    status = current_values.State
    status = KBIO.PROG_STATE(status).name

    tech_name = TECH_ID(data_info.TechniqueID).name

    # synthetic info for current record
    info = {
        "tb": current_values.TimeBase,
        "ix": data_info.TechniqueIndex,
        "tech": tech_name,
        "proc": data_info.ProcessIndex,
        "loop": data_info.loop,
        "skip": data_info.IRQskipped,
    }

    print("> data info :")
    print(info)

    ix = 0

    for _ in range(data_info.NbRows):

        if tech_name == "OCV":

            # progress through record
            inx = ix + data_info.NbCols

            # extract timestamp and one row
            t_high, t_low, *row = data_record[ix:inx]

            nb_words = len(row)
            if nb_words == 1:
                vmp3 = False
            elif nb_words == 2:
                vmp3 = True
            else:
                raise RuntimeError(
                    f"{tech_name} : unexpected record length ({nb_words})"
                )

            # compute timestamp in seconds
            t_rel = (t_high << 32) + t_low
            t = current_values.TimeBase * t_rel

            # Ewe is a float
            Ewe = api.ConvertNumericIntoSingle(row[0])

            parsed_row = {"t": t, "Ewe": Ewe}

            if vmp3:
                # Ece is a float
                Ece = api.ConvertNumericIntoSingle(row[1])
                parsed_row["Ece"] = Ece

        elif tech_name == "CP":

            inx = ix + data_info.NbCols
            t_high, t_low, *row = data_record[ix:inx]

            nb_words = len(row)
            if nb_words != 3:
                raise RuntimeError(
                    f"{tech_name} : unexpected record length ({nb_words})"
                )

            # Ewe is a float
            Ewe = api.ConvertNumericIntoSingle(row[0])

            # current is a float
            I = api.ConvertNumericIntoSingle(row[1])

            # technique cycle is an integer
            cycle = row[2]

            # compute timestamp in seconds
            t_rel = (t_high << 32) + t_low
            t = current_values.TimeBase * t_rel

            parsed_row = {"t": t, "Ewe": Ewe, "I": I, "cycle": cycle}

        else:

            # besides the previous 2 known techniques, this is provided
            # to show a raw dump of the record
            inx = ix + data_info.NbCols
            row = data_record[ix:inx]
            parsed_row = [f"0x{word:08X}" for word in row]

        print("> data record :")
        print(parsed_row)

        ix = inx

    return status



import logging
import random
import time
import multiprocessing

from datetime import datetime, timezone


def _dummy_process(
    queue: multiprocessing.Queue,
    tech: str = "random",
    delay: int = 1,
    t: int = 10,
) -> None:
    ts = te = time.perf_counter()
    nd = 0
    while te - ts < t:
        if queue.empty():
            queue.put(None)
        if te >= ts + nd * delay:
            nd += 1
            data = {
                "time": te - ts,
                "value": random.random() if tech == "random" else nd,
            }
            queue.put(data)
        time.sleep(1e-3)
        te = time.perf_counter()
    return


def get_status(
    address: str,
    channel: int,
    jobqueue: multiprocessing.Queue,
    logger: logging.Logger,
    **kwargs: dict,
) -> tuple[float, dict]:
    """
    Get the current status of the device.

    Parameters
    ----------
    address
        Not used

    channel
        Numeric, 1-indexed ID of the channel.

    dllpath
        Path to the BioLogic DLL file.

    Returns
    -------
    timestamp, ready, metadata: tuple[float, bool, dict]
        Returns a tuple containing the timestamp, readiness status, and
        associated metadata.

    """
    logger.debug("in 'dummy.get_status'")
    dt = datetime.now(timezone.utc)
    metadata = {"address": address, "channel": channel}
    ready = jobqueue.empty()
    return dt.timestamp(), ready, metadata


def get_data(
    address: str,
    channel: int,
    jobqueue: multiprocessing.Queue,
    logger: logging.Logger,
    **kwargs: dict,
) -> tuple[float, dict]:
    """
    Get cached data from the device.

    Parameters
    ----------
    address
        IP address of the potentiostat.

    channel
        Numeric, 1-indexed ID of the channel.

    dllpath
        Path to the BioLogic DLL file.

    Returns
    -------
    timestamp, nrows, data: tuple[float, int, dict]
        Returns a tuple containing the timestamp and associated metadata.

    """
    logger.debug(
        f"in 'dummy.get_data', jobqueue is{'' if jobqueue.empty() else ' not'} empty"
    )
    dt = datetime.now(timezone.utc)
    points = []
    while not jobqueue.empty():
        v = jobqueue.get()
        if isinstance(v, dict):
            points.append(v)
    if jobqueue.empty() and len(points) > 0:
        jobqueue.put(None)
    npoints = len(points)
    data = {"data": points, "current": None}
    return dt.timestamp(), npoints, data


def start_job(
    address: str,
    channel: int,
    jobqueue: multiprocessing.Queue,
    logger: logging.Logger,
    payload: list[dict],
    **kwargs: dict,
) -> float:
    """
    Start a job on the device.

    Parameters
    ----------
    address
        IP address of the potentiostat.

    channel
        Numeric, 1-indexed ID of the channel.

    payload
        A protocol describing the techniques to be executed and their order.

    capacity
        The capacity information for the studied battery cell. Only required for
        battery-testing applications or for payloads where currents are specified
        using C or D rates.

    Returns
    -------
    timestamp
        A timestamp corresponding to the start of the job execution.

    """
    dt = datetime.now(timezone.utc)
    logger.info("in 'dummy.start_job'")
    logger.debug(f"{payload=}")
    for ip, p in enumerate(payload):
        delay = p.get("delay", 1)
        t = p.get("time", 10)
        tech = p["technique"]
        logger.debug(
            f"starting 'dummy._dummy_process' {ip} with {tech=}, {t=}, {delay=}."
        )
        pr = multiprocessing.Process(
            target=_dummy_process, args=(jobqueue, tech, delay, t)
        )
        pr.start()
    # Delay before quitting so that processes get a chance to start
    time.sleep(1)
    return dt.timestamp()


def stop_job(
    address: str,
    channel: int,
    jobqueue: multiprocessing.Queue,
    logger: multiprocessing.Queue,
    **kwargs: dict,
) -> float:
    """
    Stop a job running on the device.

    This function stops any currently running technique on the specified channel
    of the device. No data is returned.

    Parameters
    ----------
    address
        IP address of the potentiostat.

    channel
        Numeric, 1-indexed ID of the channel.

    dllpath
        Path to the BioLogic DLL file.

    Returns
    -------
    timestamp
        A timestamp corresponding to the start of the job execution.

    """
    jobqueue.close()
    dt = datetime.now(timezone.utc)
    return dt.timestamp()



"""
Main module - executables for tomato.

"""
import argparse
import logging
import psutil
import os
from importlib import metadata

from . import daemon
from . import dbhandler
from . import setlib
from . import ketchup

log = logging.getLogger(__name__)


def _logging_setup(args):
    loglevel = min(max(30 + 10 * (args.quiet - args.verbose), 10), 50)
    logging.basicConfig(level=loglevel)
    log.debug(f"loglevel set to '{logging._levelToName[loglevel]}'")


def _default_parsers() -> tuple[argparse.ArgumentParser, argparse.ArgumentParser]:
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "--version",
        action="version",
        version=f'%(prog)s version {metadata.version("tomato")}',
    )
    parser.add_argument(
        "-t",
        "--test",
        action="store_true",
        default=False,
        help="Launch tomato in test mode.",
    )

    verbose = argparse.ArgumentParser(add_help=False)
    for p in [parser, verbose]:
        p.add_argument(
            "-v",
            "--verbose",
            action="count",
            default=0,
            help="Increase verbosity by one level.",
        )
        p.add_argument(
            "-q",
            "--quiet",
            action="count",
            default=0,
            help="Decrease verbosity by one level.",
        )
    return parser, verbose


def sync_pipelines_to_state(
    pipelines: list,
    dbpath: str,
    type: str = "sqlite3",
) -> None:
    pstate = dbhandler.pipeline_get_all(dbpath, type)
    for pip in pipelines:
        log.debug(f"checking presence of pipeline '{pip['name']}' in 'state'")
        if pip["name"] not in pstate:
            dbhandler.pipeline_insert(dbpath, pip["name"], type)
    pnames = [p["name"] for p in pipelines]
    for pname in pstate:
        if pname not in pnames:
            dbhandler.pipeline_remove(dbpath, pname, type)
    pstate = dbhandler.pipeline_get_all(dbpath, type)


def run_tomato():
    parser, _ = _default_parsers()
    args = parser.parse_args()
    _logging_setup(args)

    ppid = os.getppid()  # On Windows, tomato.exe is the parent of os.getpid()
    toms = [
        p.pid for p in psutil.process_iter() if p.name() in {"tomato", "tomato.exe"}
    ]
    toms.pop(toms.index(ppid))
    if len(toms) > 0 and not args.test:
        logging.critical("cannot run more than one instance of 'tomato'")
        logging.info(f"'tomato' is currently running as pid {toms}")
        return

    dirs = setlib.get_dirs(args.test)
    settings = setlib.get_settings(dirs.user_config_dir, dirs.user_data_dir)
    pipelines = setlib.get_pipelines(settings["devices"]["path"])
    log.debug(f"setting up 'queue' table in '{settings['queue']['path']}'")
    dbhandler.queue_setup(settings["queue"]["path"], type=settings["queue"]["type"])
    log.debug(f"setting up 'state' table in '{settings['queue']['path']}'")
    dbhandler.state_setup(settings["state"]["path"], type=settings["state"]["type"])
    sync_pipelines_to_state(
        pipelines, settings["state"]["path"], type=settings["state"]["type"]
    )

    daemon.main_loop(settings, pipelines, test=args.test)


def run_ketchup():
    parser, verbose = _default_parsers()
    subparsers = parser.add_subparsers(dest="subcommand", required=True)

    submit = subparsers.add_parser("submit")
    submit.add_argument(
        "payload",
        help="File containing the payload to be submitted to tomato.",
        default=None,
    )
    submit.set_defaults(func=ketchup.submit)

    status = subparsers.add_parser("status")
    status.add_argument(
        "jobid",
        nargs="?",
        help=(
            "The jobid of the requested job, "
            "or 'queue' for the status of the queue,"
            "or 'state' for the status of pipelines."
        ),
        default="state",
    )
    status.set_defaults(func=ketchup.status)

    cancel = subparsers.add_parser("cancel")
    cancel.add_argument(
        "jobid", help="The jobid of the job to be cancelled.", default=None
    )
    cancel.set_defaults(func=ketchup.cancel)

    load = subparsers.add_parser("load")
    load.add_argument("sample", help="Name of the sample to be loaded.", default=None)
    load.add_argument(
        "pipeline", help="Name of the pipeline to load the sample to.", default=None
    )
    load.set_defaults(func=ketchup.load)

    eject = subparsers.add_parser("eject")
    eject.add_argument(
        "pipeline", help="Name of the pipeline to eject any sample from.", default=None
    )
    eject.set_defaults(func=ketchup.eject)

    ready = subparsers.add_parser("ready")
    ready.add_argument(
        "pipeline", help="Name of the pipeline to mark as ready.", default=None
    )
    ready.set_defaults(func=ketchup.ready)

    args, extras = parser.parse_known_args()
    args, extras = verbose.parse_known_args(extras, args)
    _logging_setup(args)

    if "func" in args:
        args.func(args)



_device_to_parser = {
    "dummy": "dummy",
    "biologic": "electrochem",
}


def get_yadg_preset(method: list[dict], pipeline: dict) -> dict:
    preset = {
        "metadata": {
            "version": "4.1.1",
            "provenance": {"type": "tomato"},
            "timezone": "localtime",
        },
        "steps": [],
    }

    devices = {item["tag"]: item["driver"] for item in pipeline["devices"]}
    for dev in set([item["device"] for item in method]):
        step = {
            "tag": dev,
            "parser": _device_to_parser[devices[dev]],
            "input": {"folders": ["."], "prefix": dev, "suffix": "data.json"},
            "parameters": {"filetype": "tomato.json"},
            "externaldate": {
                "using": {
                    "file": {
                        "type": "json",
                        "path": f"{dev}_status.json",
                        "match": "uts",
                    }
                },
            },
        }
        preset["steps"].append(step)
    return preset



"""
**ketchup**: command line interface to **tomato**
-------------------------------------------------
.. codeauthor:: 
    Peter Kraus

Module of functions to interact with tomato. Includes job management functions:

- :func:`.submit` to submit a *job* to *queue*
- :func:`.status` to query the status of tomato's *pipelines*, its *queue*, or a *job*
- :func:`.cancel` to cancel a queued or kill a running *job*

Also includes *sample*/*pipeline* management functions:

- :func:`.load` to load a *sample* into a *pipeline*
- :func:`.eject` to remove any *sample* present in a *pipeline*
- :func:`.ready` to mark a *pipeline* as ready


"""
from .functions import submit, status, cancel, load, eject, ready



import setuptools
import versioneer
version=versioneer.get_version()
cmdclass=versioneer.get_cmdclass()

with open("README.md", "r", encoding="utf-8") as infile:
    readme = infile.read()

packagedir = "src"

setuptools.setup(
    name="tomato",
    version=version,
    cmdclass=cmdclass,
    author="Peter Kraus",
    author_email="peter@tondon.de",
    description="au-tomation without pain!",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/dgbowl/tomato",
    project_urls={
        "Bug Tracker": "https://github.com/dgbowl/tomato/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    package_dir={"": packagedir},
    packages=setuptools.find_packages(where=packagedir),
    python_requires=">=3.9",
    install_requires=[
        "appdirs>=1.4.0",
        "toml",
        "pyyaml",
        "psutil",
        "yadg>=4.1",
        "dgbowl_schemas>=105",
        "portalocker",
    ],
    extras_require={
        "testing": [
            "pytest",
        ],
        "docs": [
            "sphinx==4.5.0",
            "sphinx-rtd-theme",
            "sphinx-autodoc-typehints",
            "autodoc-pydantic"
        ]
    },
    entry_points={
        "console_scripts": [
            "tomato=tomato:run_tomato",
            "ketchup=tomato:run_ketchup",
            "tomato_job=tomato.drivers:tomato_job"
        ]
    },
)



import os
import textwrap
import toml
import yaml
import copy
import logging
import appdirs
from importlib import metadata
from datetime import datetime, timezone
from dataclasses import dataclass

log = logging.getLogger(__name__)

_VERSION = metadata.version("tomato")


@dataclass
class LocalDir:
    user_config_dir: str
    user_data_dir: str
    user_log_dir: str


def get_dirs(testmode: bool = False) -> appdirs.AppDirs:
    if testmode:
        dirs = LocalDir(
            user_config_dir=".",
            user_data_dir=".",
            user_log_dir=".",
        )
    else:
        dirs = appdirs.AppDirs("tomato", "dgbowl", version=_VERSION)
    log.debug(f"local config folder is '{dirs.user_config_dir}'")
    log.debug(f"local data folder is '{dirs.user_data_dir}'")
    log.debug(f"local log folder is '{dirs.user_log_dir}'")
    return dirs


def get_settings(configpath: str, datapath: str) -> dict:
    settingsfile = os.path.join(configpath, "settings.toml")
    if not os.path.exists(settingsfile):
        log.warning(f"config file not present. Writing defaults to '{settingsfile}'")
        defaults = textwrap.dedent(
            f"""\
            # Default settings for tomato v{_VERSION}
            # Generated on {str(datetime.now(timezone.utc))}
            [state]
            type = 'sqlite3'
            path = '{os.path.join(datapath, 'database.db')}'

            [queue]
            type = 'sqlite3'
            path = '{os.path.join(datapath, 'database.db')}'
            storage = '{os.path.join(datapath, 'Jobs')}'
            
            [samples]
            path = '{os.path.join(configpath, 'samples.yml')}'

            [devices]
            path = '{os.path.join(configpath, 'devices.yml')}'

            [drivers]
            [drivers.biologic]
            dllpath = 'C:\EC-Lab Development Package\EC-Lab Development Package\'
            """
        )
        if not os.path.exists(configpath):
            os.makedirs(configpath)
        with open(settingsfile, "w") as of:
            of.write(defaults)

    log.debug(f"loading tomato settings from '{settingsfile}'")
    settings = toml.load(settingsfile)

    return settings


def _default_pipelines() -> dict[str, dict]:
    data = {
        "devices": [
            {
                "name": "dummy_device",
                "address": None,
                "channels": [5, 10],
                "driver": "dummy",
                "capabilities": ["random", "sequential"],
                "pollrate": 1,
            }
        ],
        "pipelines": [
            {
                "name": "dummy-*",
                "devices": [
                    {"tag": "worker", "name": "dummy_device", "channel": "each"}
                ],
            }
        ],
    }
    return data


def get_pipelines(yamlpath: str) -> dict:
    log.debug(f"loading pipeline settings from '{yamlpath}'")
    try:
        with open(yamlpath, "r") as infile:
            jsdata = yaml.safe_load(infile)
    except FileNotFoundError:
        log.error(f"device settings not found. Running with a 'dummy' device.")
        jsdata = _default_pipelines()
    devices = jsdata["devices"]
    pipelines = jsdata["pipelines"]
    ret = []
    for pip in pipelines:
        if "*" in pip["name"]:
            data = {"name": pip["name"], "devices": []}
            assert len(pip["devices"]) == 1
            for ppars in pip["devices"]:
                for dpars in devices:
                    if dpars["name"] == ppars["name"]:
                        break
                dev = {k: v for k, v in dpars.items() if k != "channels"}
                dev["tag"] = ppars["tag"]
                data["devices"].append(dev)
                for ch in dpars["channels"]:
                    d = copy.deepcopy(data)
                    d["devices"][0]["channel"] = ch
                    d["name"] = d["name"].replace("*", f"{ch}")
                    ret.append(d)
        else:
            data = {"name": pip["name"], "devices": []}
            for ppars in pip["devices"]:
                for dpars in devices:
                    if dpars["name"] == ppars["name"]:
                        break
                dev = {k: v for k, v in dpars.items() if k != "channels"}
                dev["tag"] = ppars["tag"]
                if isinstance(ppars.get("channel"), int):
                    assert ppars["channel"] in dpars["channels"]
                    dev["channel"] = ppars["channel"]
                else:
                    assert "*" in pip["name"]
                data["devices"].append(dev)
            ret.append(data)
    return ret
